//////////////////////////////////////////////////////////////////////
//                              ExecModify                       
// Copyright 2014 Malcolm Stagg
// Modified Version 1.0: 06/05/2014
//
//////////////////////////////////////////////////////////////////////

/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "inc/ExecParseUtil.h"
#include "inc/IProcessorOpcode.h"
#include "inc/IExecutableParser.h"
#include "inc/ElfParser.h"
#include "inc/ArmOpcode.h"
#include "inc/X86Opcode.h"
#include "inc/Amd64Opcode.h"

// http://www.skyfree.org/linux/references/ELF_Format.pdf
// http://en.wikipedia.org/wiki/Executable_and_Linkable_Format#File_header

#define ELF_MAGIC_NUMBER 0x7F454C46

const char OSABI_STRING[13][16] = {
	{ "System V/Unk" },
	{ "HP-UX" },
	{ "NetBSD" },
	{ "Linux" },
	{ "Unknown" },
	{ "Unknown" },
	{ "Solaris" },
	{ "AIX" },
	{ "IRIX" },
	{ "FreeBSD" },
	{ "Unknown" },
	{ "Unknown" },
	{ "OpenBSD" }
};

const char TYPE_STRING[5][16] = {
	{ "Unknown" },
	{ "Relocatable" },
	{ "Executable" },
	{ "Shared" },
	{ "Core" }
};

bool ElfParser::LoadFile(const char * szFileName)
{
	m_fileElf = fopen(szFileName, "r+b");
	if (!m_fileElf)
	{
		LOG_ERROR("Cannot open file");
		return false;
	}

	if (!ReadHeaderInformation())
	{
		LOG_ERROR("ReadHeaderInformation failed");
		return false;
	}

	return true;
}

bool ElfParser::CloseFile()
{
	if (!m_fileElf)
	{
		return true;
	}

	fclose(m_fileElf);
	m_fileElf = NULL;
	return true;
}

bool ElfParser::ReadHeaderInformation() {
	first_part_elf_header firstHeader;
	offsets_elf_header_64 offsetsHeader64;
	final_part_elf_header finalHeader;

	Endianess byteOrder = unknown_endian;
	Endianess elfEndian = unknown_endian;
	bool is64 = false;

	if (!m_fileElf) 
	{
		LOG_ERROR("Invalid file");
		return false;
	}

	byteOrder = GetHostByteOrder();
	if (!byteOrder) 
	{
		LOG_ERROR("Cannot detect host byte order");
		return false;
	}

	if (byteOrder == big_endian) 
	{
		LOG_INFO("Host is big-endian");
	}
	else if (byteOrder == little_endian) 
	{
		LOG_INFO("Host is little-endian");
	}

	fread(&firstHeader, 1, sizeof(firstHeader), m_fileElf);

	ConvertEndian32(firstHeader.ei_magic, big_endian, byteOrder);

	printf("ELF magic number: 0x%08X\n", firstHeader.ei_magic);
	if (firstHeader.ei_magic != ELF_MAGIC_NUMBER) 
	{
		LOG_ERROR("Invalid ELF magic number");
		return false;
	}

	printf("ei_class: 0x%X", firstHeader.ei_class);
	if (firstHeader.ei_class == 2) 
	{
		printf(" (64-bit)\n");
		is64 = true;
	}
	else if (firstHeader.ei_class == 1) 
	{
		printf(" (32-bit)\n");
	}
	else 
	{
		LOG_ERROR("Invalid ei_class");
		return false;
	}

	printf("ei_data: 0x%X", firstHeader.ei_data);
	if (firstHeader.ei_data == 2) 
	{
		printf(" (big-endian)\n");
		elfEndian = big_endian;
	}
	else if (firstHeader.ei_class == 1) 
	{
		printf(" (little-endian)\n");
		elfEndian = little_endian;
	}
	else 
	{
		LOG_ERROR("Invalid ei_data");
		return false;
	}

	ConvertEndian16(firstHeader.e_type, elfEndian, byteOrder);
	ConvertEndian16(firstHeader.e_machine, elfEndian, byteOrder);
	ConvertEndian32(firstHeader.e_version, elfEndian, byteOrder);

	printf("ei_version: 0x%X", firstHeader.ei_version);
	if (firstHeader.ei_version == 1) 
	{
		printf(" (original)\n");
	}
	else 
	{
		printf(" (unknown)\n");
	}

	printf("ei_osabi: 0x%X", firstHeader.ei_osabi);
	if (firstHeader.ei_osabi <= 0x0C) 
	{
		printf(" (%s)\n", OSABI_STRING[firstHeader.ei_osabi]);
	}
	else 
	{
		printf(" (unknown)\n");
	}

	printf("ei_abiversion: 0x%X\n", firstHeader.ei_abiversion);
	printf("e_type: 0x%X", firstHeader.e_type);
	if (firstHeader.e_type <= 0x05) 
	{
		printf(" (%s)\n", TYPE_STRING[firstHeader.e_type]);
	}
	else 
	{
		printf(" (unknown)\n");
	}

	printf("e_machine: 0x%X", firstHeader.e_machine);
	switch (firstHeader.e_machine)
	{
	case 0x02:
		printf(" (SPARC)\n");
		break;
	case 0x03:
		printf(" (x86)\n");
		break;
	case 0x08:
		printf(" (MIPS)\n");
		break;
	case 0x14:
		printf(" (PowerPC)\n");
		break;
	case 0x28:
		printf(" (ARM)\n");
		break;
	case 0x2A:
		printf(" (SuperH)\n");
		break;
	case 0x32:
		printf(" (IA-64)\n");
		break;
	case 0x3E:
		printf(" (x86-64)\n");
		break;
	case 0xB7:
		printf(" (AArch64)\n");
		break;
	default:
		LOG_ERROR("Unknown Architecture");
		return false;
	}

	printf("e_version: 0x%X", firstHeader.e_version);
	if (firstHeader.e_version == 1) 
	{
		printf(" (original)\n");
	}
	else 
	{
		printf(" (unknown)\n");
	}

	m_elfHeaderInfo.is64 = is64;
	m_elfHeaderInfo.elfEndian = elfEndian;
	m_elfHeaderInfo.byteOrder = byteOrder;
	ReadOffsetsElfHeader(offsetsHeader64);

	printf("e_entry: 0x%016llX\n", offsetsHeader64.e_entry);
	printf("e_phoff: 0x%016llX\n", offsetsHeader64.e_phoff);
	printf("e_shoff: 0x%016llX\n", offsetsHeader64.e_shoff);

	fread(&finalHeader, 1, sizeof(finalHeader), m_fileElf);

	ConvertEndian32(finalHeader.e_flags, elfEndian, byteOrder);
	ConvertEndian16(finalHeader.e_ehsize, elfEndian, byteOrder);
	ConvertEndian16(finalHeader.e_phentsize, elfEndian, byteOrder);
	ConvertEndian16(finalHeader.e_phnum, elfEndian, byteOrder);
	ConvertEndian16(finalHeader.e_shentsize, elfEndian, byteOrder);
	ConvertEndian16(finalHeader.e_shnum, elfEndian, byteOrder);
	ConvertEndian16(finalHeader.e_shstrndx, elfEndian, byteOrder);

	printf("e_flags: 0x%08X\n", finalHeader.e_flags);
	printf("e_ehsize: 0x%04X\n", finalHeader.e_ehsize);
	printf("e_phentsize: 0x%04X\n", finalHeader.e_phentsize);
	printf("e_phnum: 0x%04X\n", finalHeader.e_phnum);
	printf("e_shentsize: 0x%04X\n", finalHeader.e_shentsize);
	printf("e_shnum: 0x%04X\n", finalHeader.e_shnum);
	printf("e_shstrndx: 0x%04X\n", finalHeader.e_shstrndx);

	m_elfHeaderInfo.byteOrder = byteOrder;
	m_elfHeaderInfo.elfEndian = elfEndian;
	m_elfHeaderInfo.firstHeader = firstHeader;
	m_elfHeaderInfo.offsetsHeader64 = offsetsHeader64;
	m_elfHeaderInfo.finalHeader = finalHeader;
	m_elfHeaderInfo.is64 = is64;

	return true;
}

bool ElfParser::ReadSection(int iSectionNumber)
{
	if (!m_fileElf)
	{
		LOG_ERROR("Invalid file");
		return false;
	}

	if (!m_elfHeaderInfo.offsetsHeader64.e_shoff)
	{
		LOG_ERROR("No section header offset found");
		return false;
	}

	if (iSectionNumber >= m_elfHeaderInfo.finalHeader.e_shnum)
	{
		return false;
	}

	fseek(m_fileElf, m_elfHeaderInfo.offsetsHeader64.e_shoff + iSectionNumber * m_elfHeaderInfo.finalHeader.e_shentsize, SEEK_SET);
	ReadSectionHeader(m_sectionHeader64);
	
	return true;
}

bool ElfParser::GetFirstSection()
{
	if (!m_fileElf)
	{
		LOG_ERROR("Invalid file");
		return false;
	}

	if (!m_elfHeaderInfo.offsetsHeader64.e_shoff)
	{
		LOG_ERROR("No section header offset found");
		return false;
	}

	m_iSectionNumber = 0;
	if (!ReadSection(m_iSectionNumber))
	{
		LOG_ERROR("Cannot read first section");
		return false;
	}

	return true;
}

bool ElfParser::GetNextSection()
{
	m_iSectionNumber++;
	return ReadSection(m_iSectionNumber);
}

bool ElfParser::FindSection(char * szSectionName, SectionType type)
{
	SectionInformation sectionInfo;

	if (!GetFirstSection())
	{
		return false;
	}

	do
	{
		if (!GetSectionInformation(sectionInfo))
		{
			return false;
		}

		if ((!szSectionName || strcmp(szSectionName, sectionInfo.szName) == 0) && 
			(type == SectionType_Unknown || type == sectionInfo.type))
		{
			return true;
		}
	} while (GetNextSection());
	
	return false;
}

bool ElfParser::PrintSectionInformation()
{
	char szName[256];

	GetString(szName, sizeof(szName), m_elfHeaderInfo.finalHeader.e_shstrndx, m_sectionHeader64.sh_name);

	printf("sh_name: [%s]\n", szName);
	printf("sh_type: 0x%08X\n", m_sectionHeader64.sh_type);
	printf("sh_flags: 0x%016llX\n", m_sectionHeader64.sh_flags);
	printf("sh_addr: 0x%016llX\n", m_sectionHeader64.sh_addr);
	printf("sh_offset: 0x%016llX\n", m_sectionHeader64.sh_offset);
	printf("sh_size: 0x%016llX\n", m_sectionHeader64.sh_size);
	printf("sh_link: 0x%08X\n", m_sectionHeader64.sh_link);
	printf("sh_info: 0x%08X\n", m_sectionHeader64.sh_info);
	printf("sh_addralign: 0x%016llX\n", m_sectionHeader64.sh_addralign);

	return true;
}

bool ElfParser::GetSectionInformation(SectionInformation &sectionInfo)
{
	GetString(sectionInfo.szName, sizeof(sectionInfo.szName), m_elfHeaderInfo.finalHeader.e_shstrndx, m_sectionHeader64.sh_name);
	switch (m_sectionHeader64.sh_type)
	{
	case 0:
		sectionInfo.type = SectionType_None;
		break;
	case 1:
		sectionInfo.type = SectionType_ProgramBits;
		break;
	case 2:
		sectionInfo.type = SectionType_SymbolTable;
		break;
	case 3:
		sectionInfo.type = SectionType_StringTable;
		break;
	case 4:
		sectionInfo.type = SectionType_Rela;
		break;
	case 5:
		sectionInfo.type = SectionType_Hash;
		break;
	case 6:
		sectionInfo.type = SectionType_Dynamic;
		break;
	case 7:
		sectionInfo.type = SectionType_Note;
		break;
	case 8:
		sectionInfo.type = SectionType_NoBits;
		break;
	case 9:
		sectionInfo.type = SectionType_Reloc;
		break;
	case 10:
		sectionInfo.type = SectionType_ShLib;
		break;
	case 11:
		sectionInfo.type = SectionType_DynSym;
		break;
	case 0x70000000:
		sectionInfo.type = SectionType_LoProc;
		break;
	case 0x7FFFFFFF:
		sectionInfo.type = SectionType_HiProc;
		break;
	case 0x80000000:
		sectionInfo.type = SectionType_LoUser;
		break;
	case 0xFFFFFFFF:
		sectionInfo.type = SectionType_HiUser;
		break;
	default:
		sectionInfo.type = SectionType_Unknown;
		break;
	}
	sectionInfo.uiOffset = m_sectionHeader64.sh_offset;
	sectionInfo.uiSize = m_sectionHeader64.sh_size;

	return true;
}

bool ElfParser::GetString(char * szOutput, unsigned int cbOutput, int iStringSec, int iSecOffset) {
	section_header_64 sectionHeader64;

	if (!m_fileElf)
	{
		LOG_ERROR("Invalid file");
		return false;
	}

	if (!szOutput || !cbOutput)
	{
		LOG_ERROR("Invalid parameters");
		return false;
	}

	szOutput[0] = 0;

	if (!m_elfHeaderInfo.offsetsHeader64.e_shoff || !iSecOffset)
	{
		return false;
	}

	fseek(m_fileElf, m_elfHeaderInfo.offsetsHeader64.e_shoff + iStringSec * m_elfHeaderInfo.finalHeader.e_shentsize, SEEK_SET);

	ReadSectionHeader(sectionHeader64);
	fseek(m_fileElf, sectionHeader64.sh_offset + iSecOffset, SEEK_SET);
	fread(szOutput, 1, cbOutput, m_fileElf);
	szOutput[cbOutput - 1] = NULL;

	return true;
}

bool ElfParser::ReadSymbol(uint64_t addr)
{
	if (!m_fileElf)
	{
		LOG_ERROR("Invalid file");
		return false;
	}

	if (!m_sectionHeader64.sh_offset)
	{
		LOG_ERROR("No section header offset found");
		return false;
	}

	if (addr < m_sectionHeader64.sh_offset || addr >= m_sectionHeader64.sh_offset + m_sectionHeader64.sh_size) 
	{
		return false;
	}

	fseek(m_fileElf, addr, SEEK_SET);
	ReadSymbolHeader(m_symbol64);
	return true;
}

bool ElfParser::GetFirstSymbol()
{
	m_uiSymbolAddress = m_sectionHeader64.sh_offset;
	return ReadSymbol(m_uiSymbolAddress);
}

bool ElfParser::GetNextSymbol()
{
	if (!m_elfHeaderInfo.is64)
	{
		m_uiSymbolAddress += sizeof(symbol_32);
	}
	else
	{
		m_uiSymbolAddress += sizeof(symbol_64);
	}
	return ReadSymbol(m_uiSymbolAddress);
}

bool ElfParser::FindSymbol(char * szSymbolName, SymbolType type)
{
	SymbolInformation symbolInfo;

	if (!GetFirstSymbol())
	{
		return false;
	}

	do
	{
		if (!GetSymbolInformation(symbolInfo))
		{
			return false;
		}

		if ((!szSymbolName || strcmp(szSymbolName, symbolInfo.szName) == 0) &&
			(type == SymbolType_Unknown || type == symbolInfo.type))
		{
			return true;
		}

	} while (GetNextSymbol());

	return false;
}

bool ElfParser::PrintSymbolInformation()
{
	SymbolInformation symbolInfo;

	if (!GetSymbolInformation(symbolInfo))
	{
		return false;
	}

	printf("[%s] (%s %s) @ 0x%016llX %lluB sec 0x%X\n", symbolInfo.szName, SymbolBindingString[symbolInfo.binding], SymbolTypeString[symbolInfo.type], m_symbol64.st_value, m_symbol64.st_size, m_symbol64.st_shndx);

	return true;
}

bool ElfParser::GetSymbolInformation(SymbolInformation &symbolInfo)
{
	const char * szSymBind = SymbolBindingString[SymbolBinding_Unknown];
	const char * szSymType = SymbolTypeString[SymbolType_Unknown];
	int iSymBind = 0, iSymType = 0;

	if (!m_fileElf)
	{
		LOG_ERROR("Cannot open file");
		return false;
	}

	GetString(symbolInfo.szName, sizeof(symbolInfo.szName), m_sectionHeader64.sh_link, m_symbol64.st_name);
	iSymBind = m_symbol64.st_info >> 4;
	iSymType = m_symbol64.st_info & 0xF;

	switch (iSymBind)
	{
	case 0:
		symbolInfo.binding = SymbolBinding_Local;
		break;
	case 1:
		symbolInfo.binding = SymbolBinding_Global;
		break;
	case 2:
		symbolInfo.binding = SymbolBinding_Weak;
		break;
	case 13:
		symbolInfo.binding = SymbolBinding_LoProc;
		break;
	case 15:
		symbolInfo.binding = SymbolBinding_HiProc;
		break;
	default:
		symbolInfo.binding = SymbolBinding_Unknown;
		break;
	}

	switch (iSymType)
	{
	case 0:
		symbolInfo.type = SymbolType_None;
		break;
	case 1:
		symbolInfo.type = SymbolType_Object;
		break;
	case 2:
		symbolInfo.type = SymbolType_Function;
		break;
	case 3:
		symbolInfo.type = SymbolType_Section;
		break;
	case 4:
		symbolInfo.type = SymbolType_File;
		break;
	case 14:
		symbolInfo.type = SymbolType_LoProc;
		break;
	case 15:
		symbolInfo.type = SymbolType_HiProc;
		break;
	default:
		symbolInfo.type = SymbolType_Unknown;
		break;
	}

	symbolInfo.uiVirtualAddress = m_symbol64.st_value;
	symbolInfo.uiSize = m_symbol64.st_size;
		
	return true;
}

bool ElfParser::GetFunctionVirtualAddress(char * szSymtab, char * szFunctionName, uint64_t &uiVirtualAddress, uint64_t &uiSize)
{
	if (!FindSection(szSymtab, SectionType_SymbolTable))
	{
		LOG_ERROR("Cannot find symbol table");
		return false;
	}

	if (!FindSymbol(szFunctionName, SymbolType_Function))
	{
		LOG_ERROR("Cannot find function");
		return false;
	}

	printf("Function found at address 0x%016llX", m_symbol64.st_value);

	uiVirtualAddress = m_symbol64.st_value;
	uiSize = m_symbol64.st_size;

	return true;
}

bool ElfParser::GetEntryPointVirtualAddress(uint64_t &uiVirtualAddress)
{
	uiVirtualAddress = m_elfHeaderInfo.offsetsHeader64.e_entry;
	return true;
}

bool ElfParser::VirtualMemoryToFileOffset(uint64_t uiVirtualAddress, uint64_t &uiOffset)
{
	int i;
	program_header_64 programHeader64;

	if (!m_fileElf)
	{
		LOG_ERROR("Cannot open file");
		return false;
	}

	if (!m_elfHeaderInfo.offsetsHeader64.e_phoff)
	{
		LOG_WARNING("No program header offset found");
		uiOffset = uiVirtualAddress;
		return true;
	}

	for (i = 0; i < m_elfHeaderInfo.finalHeader.e_phnum; i++) {
		fseek(m_fileElf, m_elfHeaderInfo.offsetsHeader64.e_phoff + i * m_elfHeaderInfo.finalHeader.e_phentsize, SEEK_SET);
		ReadProgramHeader(programHeader64);

		if (programHeader64.p_vaddr <= uiVirtualAddress && programHeader64.p_vaddr + programHeader64.p_memsz >= uiVirtualAddress)
		{
			uiOffset = programHeader64.p_offset + (uiVirtualAddress - programHeader64.p_vaddr);
			return true;
		}
	}

	return false;
}

bool ElfParser::FileOffsetToVirtualMemory(uint64_t uiOffset, uint64_t &uiVirtualAddress)
{
	int i;
	program_header_64 programHeader64;

	if (!m_fileElf)
	{
		LOG_ERROR("Cannot open file");
		return false;
	}

	if (!m_elfHeaderInfo.offsetsHeader64.e_phoff)
	{
		LOG_WARNING("No program header offset found");
		uiVirtualAddress = uiOffset;
		return true;
	}

	for (i = 0; i < m_elfHeaderInfo.finalHeader.e_phnum; i++) {
		fseek(m_fileElf, m_elfHeaderInfo.offsetsHeader64.e_phoff + i * m_elfHeaderInfo.finalHeader.e_phentsize, SEEK_SET);
		ReadProgramHeader(programHeader64);

		if (programHeader64.p_offset <= uiOffset && programHeader64.p_offset + programHeader64.p_filesz >= uiOffset)
		{
			uiVirtualAddress = programHeader64.p_vaddr + (uiOffset - programHeader64.p_offset);
			return true;
		}
	}

	return false;
}

bool ElfParser::GetProcessorOpcodes(IProcessorOpcode * &pProcOpcode)
{
	if (m_elfHeaderInfo.firstHeader.e_machine == 0x28) // ARM
	{
		if (!m_pProcOpcode)
		{
			m_pProcOpcode = new ArmOpcode();
			pProcOpcode = m_pProcOpcode;
		}
		return true;
	}
	else if (m_elfHeaderInfo.firstHeader.e_machine == 0x03) // x86
	{
		if (!m_pProcOpcode)
		{
			m_pProcOpcode = new X86Opcode();
			pProcOpcode = m_pProcOpcode;
		}
		return true;
	}
	else if (m_elfHeaderInfo.firstHeader.e_machine == 0x32) // IA-64
	{
		if (!m_pProcOpcode)
		{
			m_pProcOpcode = new Amd64Opcode();
			pProcOpcode = m_pProcOpcode;
		}
		return true;
	}

	return false;
}

bool ElfParser::OverwriteData(uint64_t uiOffset, char * cData, int cbData)
{
	int i;

	if (!m_fileElf)
	{
		LOG_ERROR("Cannot open file");
		return false;
	}
	if (fseek(m_fileElf, uiOffset, SEEK_SET))
	{
		LOG_ERROR("fseek failed");
		return false;
	}
	if (fwrite(cData, 1, cbData, m_fileElf) != cbData)
	{
		LOG_ERROR("fwrite failed");
		return false;
	}

	return true;
}

bool ElfParser::ReadOffsetsElfHeader(offsets_elf_header_64 &offsetsHeader64)
{
	if (!m_elfHeaderInfo.is64)
	{
		offsets_elf_header_32 offsetsHeader32;

		fread(&offsetsHeader32, 1, sizeof(offsetsHeader32), m_fileElf);

		ConvertEndian32(offsetsHeader32.e_entry, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(offsetsHeader32.e_phoff, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(offsetsHeader32.e_shoff, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);

		offsetsHeader64.e_entry = offsetsHeader32.e_entry;
		offsetsHeader64.e_phoff = offsetsHeader32.e_phoff;
		offsetsHeader64.e_shoff = offsetsHeader32.e_shoff;
	}
	else
	{
		fread(&offsetsHeader64, 1, sizeof(offsetsHeader64), m_fileElf);

		ConvertEndian64(offsetsHeader64.e_entry, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(offsetsHeader64.e_phoff, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(offsetsHeader64.e_shoff, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
	}

	return true;
}

bool ElfParser::ReadSectionHeader(section_header_64 &sectionHeader64)
{

	if (!m_elfHeaderInfo.is64)
	{
		section_header_32 sectionHeader32;

		fread(&sectionHeader32, 1, sizeof(sectionHeader32), m_fileElf);
		ConvertEndian32(sectionHeader32.sh_name, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_type, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_flags, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_addr, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_offset, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_size, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_link, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_info, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_addralign, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader32.sh_entsize, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		sectionHeader64.sh_name = sectionHeader32.sh_name;
		sectionHeader64.sh_type = sectionHeader32.sh_type;
		sectionHeader64.sh_flags = sectionHeader32.sh_flags;
		sectionHeader64.sh_addr = sectionHeader32.sh_addr;
		sectionHeader64.sh_offset = sectionHeader32.sh_offset;
		sectionHeader64.sh_size = sectionHeader32.sh_size;
		sectionHeader64.sh_link = sectionHeader32.sh_link;
		sectionHeader64.sh_info = sectionHeader32.sh_info;
		sectionHeader64.sh_addralign = sectionHeader32.sh_addralign;
		sectionHeader64.sh_entsize = sectionHeader32.sh_entsize;
	}
	else
	{
		fread(&sectionHeader64, 1, sizeof(sectionHeader64), m_fileElf);
		ConvertEndian32(sectionHeader64.sh_name, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader64.sh_type, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(sectionHeader64.sh_flags, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(sectionHeader64.sh_addr, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(sectionHeader64.sh_offset, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(sectionHeader64.sh_size, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader64.sh_link, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(sectionHeader64.sh_info, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(sectionHeader64.sh_addralign, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(sectionHeader64.sh_entsize, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
	}

	return true;
}

bool ElfParser::ReadSymbolHeader(symbol_64 &symbol64)
{
	if (!m_elfHeaderInfo.is64)
	{
		symbol_32 symbol32;
		fread(&symbol32, 1, sizeof(symbol32), m_fileElf);

		ConvertEndian32(symbol32.st_name, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(symbol32.st_value, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(symbol32.st_size, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian16(symbol32.st_shndx, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		symbol64.st_name = symbol32.st_name;
		symbol64.st_value = symbol32.st_value;
		symbol64.st_size = symbol32.st_size;
		symbol64.st_info = symbol32.st_info;
		symbol64.st_other = symbol32.st_other;
		symbol64.st_shndx = symbol32.st_shndx;
	}
	else
	{
		fread(&symbol64, 1, sizeof(symbol64), m_fileElf);

		ConvertEndian32(symbol64.st_name, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(symbol64.st_value, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(symbol64.st_size, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian16(symbol64.st_shndx, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
	}

	return true;
}

bool ElfParser::ReadProgramHeader(program_header_64 &programHeader64)
{

	if (!m_elfHeaderInfo.is64)
	{
		program_header_32 programHeader32;

		fread(&programHeader32, 1, sizeof(programHeader32), m_fileElf);

		ConvertEndian32(programHeader32.p_type, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(programHeader32.p_offset, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(programHeader32.p_vaddr, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(programHeader32.p_paddr, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(programHeader32.p_filesz, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(programHeader32.p_memsz, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(programHeader32.p_flags, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian32(programHeader32.p_align, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		programHeader64.p_type = programHeader32.p_type;
		programHeader64.p_offset = programHeader32.p_offset;
		programHeader64.p_vaddr = programHeader32.p_vaddr;
		programHeader64.p_paddr = programHeader32.p_paddr;
		programHeader64.p_filesz = programHeader32.p_filesz;
		programHeader64.p_memsz = programHeader32.p_memsz;
		programHeader64.p_flags = programHeader32.p_flags;
		programHeader64.p_align = programHeader32.p_align;
	}
	else
	{
		fread(&programHeader64, 1, sizeof(programHeader64), m_fileElf);

		ConvertEndian32(programHeader64.p_type, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(programHeader64.p_offset, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(programHeader64.p_vaddr, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(programHeader64.p_paddr, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(programHeader64.p_filesz, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(programHeader64.p_memsz, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(programHeader64.p_flags, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
		ConvertEndian64(programHeader64.p_align, m_elfHeaderInfo.elfEndian, m_elfHeaderInfo.byteOrder);
	}

	return true;
}
