//////////////////////////////////////////////////////////////////////
//                              ExecModify                       
// Copyright 2014 Malcolm Stagg
// Version 1.0: 06/05/2014
//--------------------------------------------------------------------
// Changes:
// Version 1.0 (06/05/2014):
// - Only supports ELF executables
// - Only supports "stubbing out" a function
// - Only supports x86, AMD64, and ARM
// - Limited error checking/handling
// - No relocatable ELF files
//////////////////////////////////////////////////////////////////////

/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "inc/ExecParseUtil.h"
#include "inc/IProcessorOpcode.h"
#include "inc/IExecutableParser.h"
#include "inc/ElfParser.h"
#include "inc/ArmOpcode.h"

#define MAX_OPT_STRING 16

enum ExecutableTypes {
	Executable_Unknown = 0,
	Executable_ELF = 1,
	NumberOfExecutables
};

const char cExecutables[][MAX_OPT_STRING] = {
	{ "" },
	{ "elf" }
};

enum ProgramCommands {
	Command_Unknown = 0,
	Command_PrintSections = 1,
	Command_PrintSymbols = 2,
	Command_GetSymbol = 3,
	Command_StubOut = 4,
	Command_GetOffset = 5,
	Command_GetVirtual = 6,
	NumberOfCommands
};

const char cCommands[][MAX_OPT_STRING] = {
	{ "" },
	{ "printsections" },
	{ "printsymbols" },
	{ "getsymbol" },
	{ "stubout" },
	{ "getoffset" },
	{ "getvirtual" }
};

enum ProgramOptions {
	Option_Unknown = 0,
	Option_Offset = 1,
	Option_Function = 2,
	Option_Virtual = 3,
	Option_EntryPoint = 4,
	Option_SymTab = 5,
	NumberOfOptions
};

enum AddressBase {
	AddressBase_None = 0,
	AddressBase_FileStart = 1,
	AddressBase_Function = 2,
	AddressBase_Virtual = 3,
	AddressBase_EntryPoint = 4
};

const char cOptions[][MAX_OPT_STRING] = {
	{ "" },
	{ "offset" },
	{ "function" },
	{ "virtual" },
	{ "entrypoint" },
	{ "symtab" }
};

typedef struct {
	AddressBase base;
	int64_t iOffset;
	int64_t iVMOffset;
	const char * szFromFunction;
} AddressOffset;

typedef struct {
	ExecutableTypes exec;
	ProgramCommands cmd;
	AddressOffset offset;
	const char * szFileName;
	const char * szArgument1;
	const char * szSymTab;
} AllOptions;

bool LoadExecutable(char * szFileName, IExecutableParser * pExecParser)
{
	if (!pExecParser)
	{
		LOG_ERROR("pExecParser not defined");
		return false;
	}
	if (!pExecParser->LoadFile(szFileName))
	{
		LOG_ERROR("LoadFile failed");
		return false;
	}
	
	return true;
}

bool PrintSections(IExecutableParser * pExecParser)
{
	if (!pExecParser)
	{
		LOG_ERROR("pExecParser not defined");
		return false;
	}
	if (!pExecParser->GetFirstSection())
	{
		return true;
	}
	printf("\n=== SECTION HEADER TABLE ===\n");
	do {
		if (!pExecParser->PrintSectionInformation())
		{
			return false;
		}
		printf("-------------------------\n");
	} while (pExecParser->GetNextSection());

	return true;
}

bool PrintSymbols(IExecutableParser * pExecParser, char * szSymTab)
{
	if (!pExecParser)
	{
		LOG_ERROR("pExecParser not defined");
		return false;
	}
	if (!pExecParser->FindSection(szSymTab, SectionType_SymbolTable))
	{
		LOG_ERROR("Could not find symbol table");
		return false;
	}
	if (!pExecParser->GetFirstSymbol())
	{
		return true;
	}
	printf("\n=== SYMBOL TABLE ===\n");
	do {
		if (!pExecParser->PrintSymbolInformation())
		{
			return false;
		}
	} while (pExecParser->GetNextSymbol());

	return true;
}

bool GetSymbol(IExecutableParser * pExecParser, char * szSymbol, char * szSymTab)
{
	if (!pExecParser)
	{
		LOG_ERROR("pExecParser not defined");
		return false;
	}
	if (!pExecParser->FindSection(szSymTab, SectionType_SymbolTable))
	{
		LOG_ERROR("Could not find symbol table");
		return false;
	}
	if (!pExecParser->FindSymbol(szSymbol, SymbolType_Unknown))
	{
		return false;
	}
	if (!pExecParser->PrintSymbolInformation())
	{
		return false;
	}
	
	return true;
}

bool GetAddressOffset(ProgramOptions opt, const char * szOptionVal, AddressOffset &offset)
{
	switch (opt)
	{
	case Option_EntryPoint:
		if (offset.base != AddressBase_None && offset.base != AddressBase_FileStart)
		{
			LOG_ERROR("Conflicting memory offset specified");
			return false;
		}
		offset.base = AddressBase_EntryPoint;
		break;
	case Option_Offset:
		if (offset.base == AddressBase_None)
		{
			offset.base = AddressBase_FileStart;
		}
		offset.iOffset = GetNumber(szOptionVal);
		break;
	case Option_Function:
		if (offset.base != AddressBase_None && offset.base != AddressBase_FileStart)
		{
			LOG_ERROR("Conflicting memory offset specified");
			return false;
		}
		offset.base = AddressBase_Function;
		offset.szFromFunction = szOptionVal;
		break;
	case Option_Virtual:
		if (offset.base != AddressBase_None && offset.base != AddressBase_FileStart)
		{
			LOG_ERROR("Conflicting memory offset specified");
			return false;
		}
		offset.base = AddressBase_Virtual;
		offset.iVMOffset = GetNumber(szOptionVal);
		break;
	}

	return true;
}

int ParseArg(const char * szValue, const char szOptions[][MAX_OPT_STRING], int iNumOptions)
{
	for (int i = 0; i < iNumOptions; i++)
	{
		if (strcmp(szValue, szOptions[i]) == 0)
		{
			return i;
		}
	}
	return 0;
}

bool ParseArgs(AllOptions &opts, int argc, const char ** argv)
{
	const int iProgramArg = 0;
	const int iExecFormat = 1;
	const int iExecFile = 2;
	const int iCommandArg = 3;

	const char * szArgument1 = NULL;
	const char * szOptionVal = NULL;
	const char * szSymTab = NULL;
	
	ProgramCommands cmd = Command_Unknown;
	ProgramOptions opt = Option_Unknown;
	ExecutableTypes exec = Executable_Unknown;
	AddressOffset offset;

	offset.base = AddressBase_None;
	offset.iOffset = 0LL;
	offset.iVMOffset = 0LL;
	offset.szFromFunction = NULL;

	for (int i = iExecFormat; i < argc; i++)
	{
		if (argv[i][0] != '-' || argv[i][1] != '-')
		{
			LOG_ERROR("Invalid argument");
			return false;
		}
		if (i == iExecFormat)
		{
			exec = (ExecutableTypes)ParseArg(argv[i] + 2, cExecutables, NumberOfExecutables);

			if (exec != Executable_ELF)
			{
				LOG_ERROR("Sorry, only ELF currently supported");
				return false;
			}

			i++; //skip file name
		}
		else if (i == iCommandArg)
		{
			cmd = (ProgramCommands)ParseArg(argv[i] + 2, cCommands, NumberOfCommands);
			
			switch (cmd)
			{
				case Command_PrintSections:
				case Command_PrintSymbols:
				case Command_StubOut:
				case Command_GetOffset:
				case Command_GetVirtual:
					break;
				case Command_GetSymbol:
					if (++i >= argc)
					{
						LOG_ERROR("Too few arguments");
						return false;
					}
					szArgument1 = argv[i];
					break;
				default:
					LOG_ERROR("Invalid Command");
					break;
			}
		}
		else
		{
			opt = (ProgramOptions)ParseArg(argv[i] + 2, cOptions, NumberOfOptions);

			switch (opt)
			{
				case Option_EntryPoint:
					break;
				case Option_Offset:
				case Option_Function:
				case Option_Virtual:
				case Option_SymTab:
					if (++i >= argc)
					{
						LOG_ERROR("Too few arguments");
						return false;
					}
					szOptionVal = argv[i];
					break;
				default:
					LOG_ERROR("Invalid Option");
					return false;
					break;
			}

			if (!GetAddressOffset(opt, szOptionVal, offset))
			{
				return false;
			}
			
			if (opt == Option_SymTab)
			{
				szSymTab = szOptionVal;
			}
		}
	}

	opts.cmd = cmd;
	opts.exec = exec;
	opts.offset = offset;
	opts.szFileName = argv[iExecFile];
	opts.szArgument1 = szArgument1;
	opts.szSymTab = szSymTab;

	return true;
}

int main(int argc, const char ** argv) {
	AllOptions opts;
	IExecutableParser * pExecParser = new ElfParser();
	IProcessorOpcode * pProcOpcode = NULL;
	uint64_t file_offset = 0ULL;
	uint64_t virtual_offset = 0ULL;
	uint64_t mem_sz = 0LL;
	char cOpcodeData[256];
	int cbOpcodeData = 0;

	if (argc < 4) {
		printf("Usage: %s <file type> <executable file> <command> [options]\n", argv[0]);
		printf("\nFile Type:\n");
		printf("  --elf\n");
		printf("      ELF file format\n");
		/*printf("  --pe\n");
		printf("      PE (EXE/DLL) file format\n");*/
		printf("\nCommands:\n");
		printf("  --printsections\n");
		printf("      Print the raw section headers\n");
		printf("  --printsymbols\n");
		printf("      Print symbol table\n");
		printf("  --getsymbol <symbol name>\n");
		printf("      Find the specified symbol\n");
		printf("  --stubout <name/address>\n");
		printf("      Stub out the function specified by name or address\n");
		printf("  --getoffset <name/address>\n");
		printf("      Get the file offset of this address\n");
		printf("  --getvirtual <name/address>\n");
		printf("      Get the virtual memory offset of this address\n");
		/*printf("  --dumphex <name/address> [--length bytes]\n");
		printf("  --dumpbin <name/address> [--length bytes]\n");
		printf("      Dump elf data to stdout. If not specified, length is\n");
		printf("      the symbol length (when available).\n");
		printf("  --modhex <name/address> --length bytes\n");
		printf("  --modbin <name/address> --length bytes\n");
		printf("      Overwrite elf data from stdin.\n");
		printf("  --redirecthex <name/address> --length bytes\n");
		printf("  --redirectbin <name/address> --length bytes\n");
		printf("      Redirect to custom code from stdin at specified location.\n");*/
		printf("\nName/Address:\n");
		printf("  --offset <offset bytes>\n");
		printf("      Use this offset from file start (or from another address)\n");
		printf("  --function <function name>\n");
		printf("      Use the address of this function. May be used along with --offset.\n");
		printf("  --virtual <virtual memory offset bytes>\n");
		printf("      Use this virtual memory offset. May be used along with --offset.\n");
		printf("  --entrypoint\n");
		printf("      Use the entry point address. May be used along with --offset.\n");
		printf("\nOther options:\n");
		printf("  --symtab <symtab name>\n");
		printf("      Use an alternate symbol table\n");
		/*printf("  --thumb\n");
		printf("      Assume ARM thumb instructions\n");*/
		/*printf("  --retval <value>\n");
		printf("      Return with this value (for stubout)\n");*/
		printf("\nNumbers can be hex (begin with 0x) or decimal.\n");
		exit(0);
	}

	if (!ParseArgs(opts, argc, argv))
	{
		LOG_ERROR("Cannot parse arguments");
		delete pExecParser;
		exit(1);
	}

	if (!LoadExecutable((char *)opts.szFileName, pExecParser))
	{
		LOG_ERROR("LoadExecutable failed");
		delete pExecParser;
		exit(1);
	}

	switch (opts.offset.base)
	{
	case AddressBase_FileStart:
		file_offset = 0;
		break;
	case AddressBase_EntryPoint:
		if (!pExecParser->GetEntryPointVirtualAddress(virtual_offset))
		{
			LOG_ERROR("Cannot get entry point virtual address");
			delete pExecParser;
			exit(1);
		}
		if (!pExecParser->VirtualMemoryToFileOffset(virtual_offset, file_offset))
		{
			LOG_ERROR("Cannot convert to file offset");
			delete pExecParser;
			exit(1);
		}
		break;
	case AddressBase_Function:
		if (!pExecParser->GetFunctionVirtualAddress((char *)opts.szSymTab, (char *)opts.offset.szFromFunction, virtual_offset, mem_sz))
		{
			LOG_ERROR("Cannot get function virtual address");
			delete pExecParser;
			exit(1);
		}
		if (!pExecParser->VirtualMemoryToFileOffset(virtual_offset, file_offset))
		{
			LOG_ERROR("Cannot convert to file offset");
			delete pExecParser;
			exit(1);
		}
		break;
	case AddressBase_Virtual:
		if (!pExecParser->VirtualMemoryToFileOffset(opts.offset.iVMOffset, file_offset))
		{
			LOG_ERROR("Cannot convert to file offset");
			delete pExecParser;
			exit(1);
		}
		break;
	}

	file_offset += opts.offset.iOffset;
	if (mem_sz >= opts.offset.iOffset)
	{
		mem_sz -= opts.offset.iOffset;
	}
	else
	{
		mem_sz = 0;
	}

	switch (opts.cmd)
	{
	case Command_PrintSections:
		if (!PrintSections(pExecParser))
		{
			LOG_ERROR("PrintSections failed");
			delete pExecParser;
			exit(1);
		}
		break;
	case Command_PrintSymbols:
		if (!PrintSymbols(pExecParser, (char *)opts.szSymTab))
		{
			LOG_ERROR("PrintSymbols failed");
			delete pExecParser;
			exit(1);
		}
		break;
	case Command_GetSymbol:
		if (!GetSymbol(pExecParser, (char *)opts.szArgument1, (char *)opts.szSymTab))
		{
			LOG_ERROR("GetSymbol failed");
			delete pExecParser;
			exit(1);
		}
		break;
	case Command_StubOut:
		if (!pExecParser->GetProcessorOpcodes(pProcOpcode))
		{
			LOG_ERROR("Cannot convert to file offset");
			delete pExecParser;
			exit(1);
		}
		cbOpcodeData = sizeof(cOpcodeData);
		if (!pProcOpcode->GetOpcode("return;", cOpcodeData, cbOpcodeData))
		{
			LOG_ERROR("Cannot get return opcode");
			delete pExecParser;
			exit(1);
		}
		if (!pExecParser->OverwriteData(file_offset, cOpcodeData, cbOpcodeData))
		{
			LOG_ERROR("Cannot overwrite file data");
			delete pExecParser;
			exit(1);
		}

		printf("\n\nCode stubbed out successfully!\n");
		break;
	case Command_GetOffset:
		printf("\n\nFile Offset = 0x%016llX, Size = 0x%016llX\n", file_offset, mem_sz);
		break;
	case Command_GetVirtual:
		if (!pExecParser->FileOffsetToVirtualMemory(file_offset, virtual_offset))
		{
			LOG_ERROR("Cannot convert to file offset");
			delete pExecParser;
			exit(1);
		}
		printf("\n\nVM Offset = 0x%016llX, Size = 0x%016llX\n", virtual_offset, mem_sz);
		break;
	}

	/*
	uint64_t va, sz, offset;
	pExecParser->GetFunctionVirtualAddress(NULL, "libSegFaultConstructor", va, sz);
	pExecParser->VirtualMemoryToFileOffset(va, offset);
	*/

	delete pExecParser;

}