//////////////////////////////////////////////////////////////////////
//                              ExecModify                       
// Copyright 2014 Malcolm Stagg
// Modified Version 1.0: 06/05/2014
//
//////////////////////////////////////////////////////////////////////

/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<http://www.gnu.org/licenses/>.
*/

#ifndef __ELF_PARSER_H__
#define __ELF_PARSER_H__

#pragma once

typedef struct 
{
	uint32_t ei_magic;
	uint8_t ei_class;
	uint8_t ei_data;
	uint8_t ei_version;
	uint8_t ei_osabi;
	uint8_t ei_abiversion;
	uint8_t ei_padding[7];
	uint16_t e_type;
	uint16_t e_machine;
	uint32_t e_version;
} first_part_elf_header;

typedef struct 
{
	uint32_t e_entry;
	uint32_t e_phoff;
	uint32_t e_shoff;
} offsets_elf_header_32;

typedef struct
{
	uint64_t e_entry;
	uint64_t e_phoff;
	uint64_t e_shoff;
} offsets_elf_header_64;

typedef struct 
{
	uint32_t e_flags;
	uint16_t e_ehsize;
	uint16_t e_phentsize;
	uint16_t e_phnum;
	uint16_t e_shentsize;
	uint16_t e_shnum;
	uint16_t e_shstrndx;
} final_part_elf_header;

typedef struct 
{
	uint32_t sh_name;
	uint32_t sh_type;
	uint32_t sh_flags;
	uint32_t sh_addr;
	uint32_t sh_offset;
	uint32_t sh_size;
	uint32_t sh_link;
	uint32_t sh_info;
	uint32_t sh_addralign;
	uint32_t sh_entsize;
} section_header_32;

typedef struct
{
	uint32_t sh_name;
	uint32_t sh_type;
	uint64_t sh_flags;
	uint64_t sh_addr;
	uint64_t sh_offset;
	uint64_t sh_size;
	uint32_t sh_link;
	uint32_t sh_info;
	uint64_t sh_addralign;
	uint64_t sh_entsize;
} section_header_64;

typedef struct 
{
	uint32_t st_name;
	uint32_t st_value;
	uint32_t st_size;
	uint8_t st_info;
	uint8_t st_other;
	uint16_t st_shndx;
} symbol_32;

typedef struct
{
	uint32_t st_name;
	uint64_t st_value;
	uint64_t st_size;
	uint8_t st_info;
	uint8_t st_other;
	uint16_t st_shndx;
} symbol_64;

typedef struct 
{
	uint32_t p_type;
	uint32_t p_offset;
	uint32_t p_vaddr;
	uint32_t p_paddr;
	uint32_t p_filesz;
	uint32_t p_memsz;
	uint32_t p_flags;
	uint32_t p_align;
} program_header_32;

typedef struct
{
	uint32_t p_type;
	uint64_t p_offset;
	uint64_t p_vaddr;
	uint64_t p_paddr;
	uint64_t p_filesz;
	uint64_t p_memsz;
	uint64_t p_flags;
	uint64_t p_align;
} program_header_64;

typedef struct {
	first_part_elf_header firstHeader;
	offsets_elf_header_64 offsetsHeader64;
	final_part_elf_header finalHeader;

	Endianess byteOrder;
	Endianess elfEndian;
	bool is64;
} header_information;

class ElfParser
	: public IExecutableParser
{
public:
	ElfParser()
		: m_fileElf(NULL)
		, m_iSectionNumber(0)
		, m_uiSymbolAddress(0)
		, m_pProcOpcode(NULL)
	{}
	~ElfParser()
	{
		CloseFile();
		if (m_pProcOpcode)
		{
			delete m_pProcOpcode;
			m_pProcOpcode = NULL;
		}
	}

	//
	// IExecutableParser
	//
	bool LoadFile(const char * szFileName);
	bool CloseFile();
	bool GetFirstSection();
	bool GetNextSection();
	bool FindSection(char * szSectionName, SectionType type);
	bool PrintSectionInformation();
	bool GetSectionInformation(SectionInformation &sectionInfo);
	bool GetFirstSymbol();
	bool GetNextSymbol();
	bool FindSymbol(char * szSymbolName, SymbolType type);
	bool PrintSymbolInformation();
	bool GetSymbolInformation(SymbolInformation &symbolInfo);
	bool GetFunctionVirtualAddress(char * szSymtab, char * szFunctionName, uint64_t &uiVirtualAddress, uint64_t &uiSize);
	bool GetEntryPointVirtualAddress(uint64_t &uiVirtualAddress);
	bool VirtualMemoryToFileOffset(uint64_t uiVirtualAddress, uint64_t &uiOffset);
	bool FileOffsetToVirtualMemory(uint64_t uiOffset, uint64_t &uiVirtualAddress);
	bool GetProcessorOpcodes(IProcessorOpcode * &pProcOpcode);
	bool OverwriteData(uint64_t uiOffset, char * cData, int cbData);

protected:
	header_information m_elfHeaderInfo;
	section_header_64 m_sectionHeader64;
	symbol_64 m_symbol64;
	FILE * m_fileElf;
	int m_iSectionNumber;
	uint64_t m_uiSymbolAddress;
	IProcessorOpcode * m_pProcOpcode;

	bool ReadHeaderInformation();
	bool ReadSection(int iSectionNumber);
	bool ReadSymbol(uint64_t addr);
	bool GetString(char * szOutput, unsigned int cbOutput, int iStringSec, int iSecOffset);
	bool ReadOffsetsElfHeader(offsets_elf_header_64 &offsetsHeader64);
	bool ReadSectionHeader(section_header_64 &sectionHeader64);
	bool ReadSymbolHeader(symbol_64 &symbol64);
	bool ReadProgramHeader(program_header_64 &programHeader64);
};

#endif